#ifndef ___NUMERIC_HPP___
#define ___NUMERIC_HPP___

#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <boost/numeric/ublas/lu.hpp>
#include <float.h>
#include <math.h>

using namespace boost::numeric;

typedef ublas::c_vector<int,2>		point;
typedef ublas::c_vector<float,2>	vec2;
typedef ublas::c_vector<float,3>	vec3;
typedef ublas::c_matrix<float,3,3>	matrix3x3;

vec2 operator * ( const vec2 & a, const vec2 & b )
{
	vec2 s;
	for( int i = 0; i < 2; ++i )
		s[i] = a[i]*b[i];
	return s;
}

template< typename T >
ublas::vector<T> m2v( const ublas::matrix<T> & m )
{
	ublas::vector<T> v( m.size1() * m.size2() );

	for( int y = 0; y < m.size2(); ++y )
		for( int x = 0; x < m.size1(); ++x )
			v[y*m.size1()+x] = m(x,y);

	return v;
}

template< typename T >
ublas::matrix<T> v2m( const ublas::vector<T> & v, int w, int h )
{
	ublas::matrix<T> m(w,h);

	for( int y = 0; y < h; ++y )
		for( int x = 0; x < w; ++x )
			m(x,y) = v(y*w+x);

	return m;
}

inline
point v2p( const vec3 & v )
{
	point p;
	vec3::value_type w = v[2];
	p[0] = v[0] / w;
	p[1] = v[1] / w;
	return p;
}

inline
vec3 p2v( const point & p )
{
	vec3 v;
	v[0] = p[0];
	v[1] = p[1];
	v[2] = 1.0;
	return v;
}

inline
matrix3x3 rotate( float rad )
{
	matrix3x3 m;
	matrix3x3::value_type c = cos(rad), s = sin(rad);
	m(0,0) =  c; m(0,1) =  s; m(0,2) = 0;
	m(1,0) = -s; m(1,1) =  c; m(1,2) = 0;
	m(2,0) =  0; m(2,1) =  0; m(2,2) = 1;
	return m;
}

inline
matrix3x3 transfer( float x, float y )
{
	matrix3x3 m;
	m(0,0) =  1; m(0,1) =  0; m(0,2) = 0;
	m(1,0) =  0; m(1,1) =  1; m(1,2) = 0;
	m(2,0) =  x; m(2,1) =  y; m(2,2) = 1;
	return m;
}

inline
matrix3x3 scaling( float w, float h )
{
	matrix3x3 m;
	m(0,0) =  w; m(0,1) =  0; m(0,2) = 0;
	m(1,0) =  0; m(1,1) =  h; m(1,2) = 0;
	m(2,0) =  0; m(2,1) =  0; m(2,2) = 1;
	return m;
}

inline 
matrix3x3 inverse( const matrix3x3 & m )
{
	matrix3x3 tmp(m);
	matrix3x3 result = ublas::identity_matrix<matrix3x3::value_type>(3);

	ublas::permutation_matrix<> pm(tmp.size1());
	ublas::lu_factorize(tmp,pm);
	ublas::lu_substitute(tmp,pm,result);

	return result;
}

#endif
