
#include <queue>
#include <iostream>
#include <vector>
#include <string>
#include <stdio.h>
#include <assert.h>
#include <utility>
#include <functional>
#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>
#include <boost/any.hpp>
#include <boost/lambda/lambda.hpp>
#include <iterator>
#include <math.h>
#include <complex>

#include "Numeric.hpp"
#include "Pixel.hpp"
#include "Image.hpp"
#include "Convert.hpp"
#include "FFT.h"

#define PI 3.1415926535897932384626433832795

GrayImagePtr LoadPGMImage( const std::string & filename );
void SavePGMImage( const std::string & filename, const GrayImage & image );

GrayImagePtr LoadBMPImage( const std::string & filename );
void SaveBMPImage( const std::string & filename, const GrayImage & image );

template< typename Image >
void SaveBMPImage( const std::string & filename, const Image & image, int width, int height );

typedef std::complex<float>		complex;
typedef ublas::matrix<complex>	frequency;

frequency encodeFFT2D( const GrayImage & image );
GrayImagePtr decodeFFT2D( frequency freq );

int main()
{
	GrayImagePtr src = LoadBMPImage("LENNA.bmp");
	GrayImagePtr msg = LoadBMPImage("msg.bmp");

	frequency src_freq = encodeFFT2D(*src);

	int width  = src->width;
	int height = src->height;

	/* Expression Template */
	SaveBMPImage( "median.bmp",      *src + median<5>(),               width, height );
	SaveBMPImage( "smooth10.bmp",    *src + Filter2(Smooth(10),10,10), width, height );
	SaveBMPImage( "bi-cubic-rotate.bmp", *src + biCubic() + affine(transfer(-128,-128)) + affine(rotate(PI/3)) + affine(transfer(128,128)), width, height );
	
	//ߍ݂Ȃ
	SaveBMPImage("direct encode-decode.bmp", *decodeFFT2D(encodeFFT2D(*src)) );

	float eps = 1e-4;

	for( int i = 0; i < 1; ++i )
	{
		const std::string encode_file = 
			std::string("images\\encode") +
			boost::lexical_cast<std::string>(eps) + ".bmp";

		const std::string decode_file = 
			std::string("images\\decode") +
			boost::lexical_cast<std::string>(eps) + ".bmp";

		//encode
		{
			frequency freq = src_freq;

			for( int y = 0; y < height; ++y )
				for( int x = 0; x < width; ++x )
				{					
					float e = (*msg)(x,y) > 127 ? eps : 0.0;
					freq(x,y) += complex(e,e);
				}

			SaveBMPImage( encode_file, *decodeFFT2D(freq) );
		}

		//decode
		{
			frequency msg = encodeFFT2D(*LoadBMPImage(encode_file)) - src_freq;

			GrayImage decode(width,height);

			for( int y = 0; y < height; ++y )
				for( int x = 0; x < width; ++x )
					if( msg(x,y).real() > 0 || msg(x,y).imag() > 0 )
							decode(x,y) = 255;
					else	decode(x,y) = 0;

			SaveBMPImage( decode_file, decode );
		}
		
		eps /= 10.0f;
	}

	fft(0, NULL, NULL);

	return 0;
}



frequency encodeFFT2D( const GrayImage & image )
{
	int width  = image.width;
	int height = image.height;
	
	frequency freq( width, height );

	std::vector<float> real(std::max(width,height));
	std::vector<float> imag(std::max(width,height));

	for( int y = 0; y < height; ++y )
	{
		for( int x = 0; x < width; ++x )
		{
			real[x] = image(x,y) / 255.0f;
			imag[x] = 0.0f;
		}

		fft(width, &real[0], &imag[0]);

		for( int x = 0; x < width; ++x )
			freq(x,y) = complex(real[x],imag[x]);
	}

	for( int x = 0; x < height; ++x )
	{
		for( int y = 0; y < width; ++y )
		{
			real[y] = freq(x,y).real();
			imag[y] = freq(x,y).imag();
		}

		fft(height, &real[0], &imag[0]);

		for( int y = 0; y < width; ++y )
			freq(x,y) = complex(real[y],imag[y]);
	}

	return freq;
}

GrayImagePtr decodeFFT2D( frequency freq )
{
	int width  = freq.size1();
	int height = freq.size2();

	GrayImagePtr image( new GrayImage(width,height) );

	std::vector<float> real(std::max(width,height));
	std::vector<float> imag(std::max(width,height));
	
	for( int x = 0; x < height; ++x )
	{
		for( int y = 0; y < width; ++y )
		{
			real[y] = freq(x,y).real();
			imag[y] = freq(x,y).imag();
		}

		fft(-height, &real[0], &imag[0]);

		for( int y = 0; y < width; ++y )
			freq(x,y) = complex(real[y],imag[y]);
	}

	for( int y = 0; y < height; ++y )
	{
		for( int x = 0; x < width; ++x )
		{
			real[x] = freq(x,y).real();
			imag[x] = freq(x,y).imag();
		}

		fft(-width, &real[0], &imag[0]);

		for( int x = 0; x < width; ++x )
			(*image)(x,y) = round<int>(real[x]*255.0f);
	}

	return image;
}

/////////////////////////////////////////////////

GrayImagePtr LoadPGMImage( const std::string & filename )
{
	FILE* fp = fopen( filename.c_str(), "rb" );
	assert( fp != NULL );
		
	char buff[1024];

	//P5check
	fgets( buff, sizeof(buff), fp );
	assert( buff[0] == 'P' && buff[1] == '5' );

	// read size
	int w= 0, h = 0;{
		while(true)
		{
			fgets( buff, sizeof(buff), fp );
			if( buff[0] == '#' ) continue;
			sscanf(buff,"%d %d", &w,&h);
			break;
		}
	}

	//max depath
	while(true)
	{
		fgets(buff,sizeof(buff),fp);
		if( buff[0] == '#' ) continue;

		int depth;
		sscanf(buff,"%d", &depth);

		assert( depth == 255 );
		break;
	}

	GrayImagePtr image( new GrayImage(w,h) );

	for( int y = 0; y < h; ++y )
		for( int x = 0; x < w; ++x )
			(*image)(x,y) = fgetc(fp);

	fclose(fp);

	return image;
}

void SavePGMImage( const std::string & filename, const GrayImage & image )
{
	FILE* fp = fopen(filename.c_str(),"wb");
	assert( fp != NULL );

	fputs("P5\n",fp);
	fputs("# Created by Image Processing\n", fp );

	fprintf(fp, "%d %d\n", image.width, image.height );
	fprintf(fp, "%d\n", 255 );

	for( int y = 0; y < image.height; ++y )
		for( int x = 0; x < image.width; ++x )
			fputc( image(x,y), fp );

	fclose(fp);
}

#include <windows.h>

//gGfBAł܂

GrayImagePtr LoadBMPImage( const std::string & filename )
{
	BITMAPFILEHEADER bmf;
	BITMAPINFOHEADER bmi;

	FILE* fp = fopen(filename.c_str(), "rb");

	assert( fp != NULL );

	fread((void*)&bmf, sizeof(BITMAPFILEHEADER), 1, fp );
	fread((void*)&bmi, sizeof(BITMAPINFOHEADER), 1, fp );

	assert( bmi.biBitCount == 8 && bmi.biCompression == 0 );
	
	GrayImagePtr image( new GrayImage(bmi.biWidth,bmi.biHeight) );

	fseek( fp, bmf.bfOffBits, SEEK_SET );

	int patting = ((bmi.biWidth+3)&~3) - bmi.biWidth;

	for( int y = bmi.biHeight-1; y >= 0; --y )
	{
		for( int x = 0; x < bmi.biWidth; ++x )
			(*image)(x,y) = fgetc(fp);
		
		for( int i = 0; i < patting; ++i )
			fgetc(fp);
	}

	fclose(fp);

	return image;
}

void SaveBMPImage( const std::string & filename, const GrayImage & image )
{
	BITMAPFILEHEADER bmf;
	BITMAPINFOHEADER bmi;

	FILE* fp = fopen(filename.c_str(), "wb");

	assert( fp != NULL );

	memset( (void*)&bmf, 0, sizeof(BITMAPFILEHEADER) );
	memset( (void*)&bmi, 0, sizeof(BITMAPINFOHEADER) );
	
	bmi.biSize		= sizeof(BITMAPINFOHEADER);
	bmi.biWidth		= image.width;
	bmi.biHeight	= image.height;
	bmi.biPlanes	= 1;
	bmi.biBitCount	= 8;
	bmi.biCompression = BI_RGB;
	bmi.biSizeImage	= image.width * image.height;

	bmf.bfOffBits = (1<<bmi.biBitCount) * 4 + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	bmf.bfType    = 'B' + ('M'<<8);
	bmf.bfSize    = bmf.bfOffBits + bmi.biSizeImage * bmi.biBitCount / 8;
	
	fwrite( (void*)&bmf, sizeof(BITMAPFILEHEADER), 1, fp );
	fwrite( (void*)&bmi, sizeof(BITMAPINFOHEADER), 1, fp );

	for( int i = 0; i < 256; ++i )
		for( int j = 0; j < 4; ++j )
			fputc(i,fp);

	int patting = ((bmi.biWidth+3)&~3) - bmi.biWidth;

	for( int y = bmi.biHeight-1; y >= 0; --y )
	{
		for( int x = 0; x < bmi.biWidth; ++x )
			fputc( image(x,y), fp );
		
		for( int i = 0; i < patting; ++i )
			fputc(0,fp);
	}

	fclose(fp);

	return;
}

template< typename Image >
void SaveBMPImage( const std::string & filename, const Image & image, int width, int height )
{
	BITMAPFILEHEADER bmf;
	BITMAPINFOHEADER bmi;

	FILE* fp = fopen(filename.c_str(), "wb");

	assert( fp != NULL );

	memset( (void*)&bmf, 0, sizeof(BITMAPFILEHEADER) );
	memset( (void*)&bmi, 0, sizeof(BITMAPINFOHEADER) );
	
	bmi.biSize		= sizeof(BITMAPINFOHEADER);
	bmi.biWidth		= width;
	bmi.biHeight	= height;
	bmi.biPlanes	= 1;
	bmi.biBitCount	= 8;
	bmi.biCompression = BI_RGB;
	bmi.biSizeImage	= bmi.biWidth * bmi.biHeight;

	bmf.bfOffBits = (1<<bmi.biBitCount) * 4 + sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	bmf.bfType    = 'B' + ('M'<<8);
	bmf.bfSize    = bmf.bfOffBits + bmi.biSizeImage * bmi.biBitCount / 8;
	
	fwrite( (void*)&bmf, sizeof(BITMAPFILEHEADER), 1, fp );
	fwrite( (void*)&bmi, sizeof(BITMAPINFOHEADER), 1, fp );

	for( int i = 0; i < 256; ++i )
		for( int j = 0; j < 4; ++j )
			fputc(i,fp);

	int patting = ((bmi.biWidth+3)&~3) - bmi.biWidth;

	for( int y = bmi.biHeight-1; y >= 0; --y )
	{
		for( int x = 0; x < bmi.biWidth; ++x )
		{
			typedef typename Image::value_type pixel;
			pixel ma = Range<pixel>::MAX;
			pixel mi = Range<pixel>::MIN;
			pixel p = image(x,y);
			fputc( roundUC(double(p-mi)/ma*255), fp );
		}
		
		for( int i = 0; i < patting; ++i )
			fputc(0,fp);
	}

	fclose(fp);

	return;	
}

